#include "file_handler.h"
#include "http_configuration.h"
#include "bgi_lib.h"


#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/resource.h>

extern "C" {
  int open(...);
}
struct file_cookie {
  char* path;
  char* root;
  int path_len;

  file_cookie(char* path,char* r) {
    path = path;
    path_len = strlen(path);
    root = strdup(r);
  }
  ~file_cookie() {
    delete root;
  }
       
};

static int allow_filename(char* filename) {
  while(*filename)  {
    if(*filename == '.' &&*(filename+1) == '.') {
      return 0;
    }
    filename++;
  }
  return 1;
}


void* file_init(char* path, char* args) {
  if(debug) {
    printf("mounting file handler on %s with args %s\n",
	   path,args);
  }
  return new file_cookie(path,current_configuration->root);
}

int file_umount(char* path, void* cookie) {
  file_cookie* fc = (file_cookie*)cookie;
  delete fc;
  return(0);
}

int file_handler(void* cookie, char* method, char* url , char* version,
		 sock_buf* buf) {
  file_cookie* fc = (file_cookie*)cookie;

  if(debug) {
    printf("file handler called to  %s on %s, using version %s\n",
	   method,url,version ? version : "0.9");
  }
  if(strcmp(method,"GET")) {
    return 501;
  }
  char* ext=0;
  char* cp = url;
  while(*cp) {
    if (*cp == '.') {
      ext = cp+1;
    }
    cp++;
  }
  char filename[1024];

  sprintf(filename,"%s/%s",fc->root,url+fc->path_len);
  if(!allow_filename(filename)) {
    return 401;
  }
  if(debug) {
    printf("trying pathname %s\n",filename);
  }
  struct stat sb;
  long page_size = sysconf(_SC_PAGESIZE);
  
  if(stat(filename,&sb) <0) {
    if(debug) {
      printf("stat error for %s\n",filename);
      perror("stat");
    }
    return(404);
  }
  
  if(sb.st_mode & S_IFDIR) {
    strcat(filename,"/index.html");
    if(!allow_filename(filename)) {
      return 401;
    }
    if(stat(filename,&sb) <0) {
      if(debug) {
	printf("stat error for %s\n",filename);
	perror("stat");
      }
      return(404);
    }
  }
  

  if(sb.st_size == 0 ) {
    return 200;
  }
  int len;
 struct sockaddr addr; 
 len = sizeof(struct sockaddr);
 struct timeval date;
 gettimeofday(&date);

 if ((getpeername(buf->sock, &addr, &len)) >= 0) {
   log_message(method,url,0,0,200,date.tv_sec,&addr,len);
 }
  int fd = open(filename,0);
  if(fd <0) {
    perror("open:");
    return 500;
  }
  if(version) {
    char* type;
    if(ext) {
      type =  ext_to_mime(ext);
    } else {
      type = "text/html";
    }
    mime_headers(buf->sock,sb.st_size,type,0,sb.st_mtime);
  }
      
  
  unsigned request_size = (sb.st_size + page_size - 1) & ~page_size;
  char* data = mmap(0,request_size,PROT_READ,MAP_SHARED,fd,0);
  if(data == 0) {
    perror("mmap:");
    return 500;
  }
  close(fd);
  off_t bytes_left= sb.st_size;
  cp = data;
  while(bytes_left) {
    int n = write(buf->sock,cp,bytes_left);
    if(n < 0) {
      goto done;
    }
    cp +=n;
    bytes_left -= n;
  }
 done:
  if(munmap(data,request_size) <0) {
    perror("munmap:");
  }

  return(200);
}
