#ifndef BGI_LIB_H
#define BGI_LIB_H

extern int debug;

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

/* This file defines library routines available for use in BGI modules */

typedef struct _sock_buf {
  char* buffer;
  int buf_size;
  char* end_of_buffer;
  char* current_ptr;
  int sock;
} sock_buf;


typedef void* (*init_func)(char* path,char* args);
typedef int (*umount_func)(char* path,void* cookie);
typedef int (*handler_func)(void* cookie, char* method, char* url,
			    char* version, sock_buf* buf);

typedef void* (*log_open_func)(char* args);
typedef int  (*log_close_func)(void* cookie);
typedef int  (*log_message_func)(void* cookie, char* method, char* url,
				char* ident, char* auth_user, int status,
				time_t date,struct sockaddr* host, int h_size);
     

typedef struct _accept {
  char* value;
  struct _accept* next;
} accept_val;

typedef struct _extra_headers {
  char* field;
  char* value;
  struct _extra_headers* next;
} extra_headers;

typedef struct _header {
  char* From;
  accept_val* Accept;
  char* Accept_Encoding;
  char* Accept_Language;
  char* User_Agent;
  char* Referer;
  char* Authorization;
  char* ChargeTo;
  extra_headers* extras;
} header;

#ifdef __cplusplus
extern "C" {
#endif
  char** split(char* string,char* pat,int collapse);
  char** split_any(char* string, char* pat, int collapse);

  header* parse_header(unsigned char* buf, int size, int* offset, int* valid, 
		       int socket);
  void http_error(sock_buf* sb, int code, char* version);
  void mime_headers(int sock, int size, char* type,char* encoding,time_t date);
  char* ext_to_mime(char* ext);

  void* create_bgi_mutex();
  int   lock_bgi_mutex(void* the_mutex);
  int   unlock_bgi_mutex(void* the_mutex);
  int   free_bgi_mutex(void* the_mutex);

  int log_message( char* method, char* url,char* ident, char*
		       auth_user, int status, time_t date,struct sockaddr* in,
		  int in_size);

#ifdef __cplusplus
}
#endif

#endif





