#ifndef LOG_H
#define LOG_H

#include "bgi_lib.h"
#include "dlist.h"

class log_handler {
 public:
  log_open_func open_func;
  log_close_func close_func;
  log_message_func mess_func;

  void* cookie;

  log_handler(log_open_func o, log_close_func c, log_message_func m) {
    open_func = o;
    close_func = c;
    mess_func = m;
  }

  log_handler(char* prefix, char* library) {
    void* dl_handle = dlopen(library,RTLD_LAZY);
    if(dl_handle == 0) {
      printf("open: %s\n",dlerror());
    }
    char buf[1024];
    sprintf(buf,"%s_open", prefix);
    open_func = (log_open_func)dlsym(dl_handle, buf);
    if(open_func == 0) {
      printf("%s: %s\n",buf,dlerror());
      dlclose(dl_handle);
    }
    sprintf(buf,"%s_close", prefix);
    close_func = (log_close_func)dlsym(dl_handle, buf);
    if(close_func == 0) {
      printf("%s: %s\n",buf,dlerror());
      open_func = 0;
      dlclose(dl_handle);
    }
    sprintf(buf,"%s_message", prefix);
    mess_func = (log_message_func)dlsym(dl_handle, buf);
    if(mess_func == 0) {
      printf("%s: %s\n",buf,dlerror());
      close_func = 0;
      open_func = 0;
      dlclose(dl_handle);
    }
  }

  void* open(char* args) {
    cookie = open_func(args);
    return cookie;
  }

  int close() {
    return  close_func(cookie);
  }
  
  int message(char* method, char* url, char*id, char* auth, int status, 
	      time_t date,struct sockaddr* host, int host_size) {
    return mess_func(cookie, method, url, id, auth, status, date,host,
		     host_size);
  }
};
 

class logger {
 public:
  dlist<log_handler>*  log_list;
  
  logger() {
    log_list = new dlist<log_handler>;
  }
  ~logger() {
    delete log_list;
  }
  int push(char* prefix, char* library,char* args) {
    log_handler* tmp = new log_handler(prefix,library);
    if(tmp->open_func == 0 ) {
      delete tmp;
      return -1;
    } else {
      if(tmp->open(args)== 0) {
	delete tmp;
	return -1;
      }
      log_list->append(0,tmp);
      return 0;
    }
  }
  int message(char* method, char* url, char*id, char* auth, int status, 
	      time_t date,struct sockaddr* host, int host_size) {
    dlist_node<log_handler>* tmp = log_list->the_list->next;
    int result=0;
    while(tmp != log_list->the_list) {
      result |= tmp->data->message(method,url,id,auth,status,date,host,
				   host_size);
      tmp = tmp->next;
    }
    return result;
  }
};
  
#endif
