#ifndef MT_LIST_H
#define MT_LIST_H

#ifdef HAVE_CONFIG_H
#include "mt_config.h"
#endif

#include "libmt.h"
#include "dlist.h"


template <class C>
class mt_dlist : public dlist<C>, public shared_object {
public:
  mt_dlist() : dlist<C>(), shared_object() { };
  // add to start of list;
  void prepend(char* key, C* object) {
    lock();
    dlist<C>::prepend(key,object);
    unlock();
  }
  // add to end of list
  void append(char* key, C* object) {
    lock();
    dlist<C>::append(key,object);
    unlock();
  }
  // add to ordered list
  int add(char* key, C* object,int no_lock=0) {
    if(!no_lock) {
      lock();
    }
    dlist<C>::add(key,object);
    if(!no_lock) {
      unlock();
    }
    return 0;
  }
  // search for key in ordered list
  C* find(char* key,int no_lock=0,C* not_found=0) {
    if(!no_lock) {
      lock();
    }
    C* result = dlist<C>::find(key,not_found);
    
    if(!no_lock) {
      unlock();
    }
    return result;
  }
  // remove key in ordered list
  void remove(char* key) {
    lock();
    dlist<C>::remove(key);
    unlock();
  }
    
  // pop the top element from the list
  C* pop() {
    lock();
    C* value = dlist<C>::pop();
    unlock();
    return value;
  }
  void map(void (*f)(char* k,C* i)) {
    lock();
    dlist<C>::map(f);
    unlock();
  }
};
#endif
