#include "bgi_lib.h"
#include "utils.h"

#ifdef HAVE_SYSENT_H
#include <sysent.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <stdlib.h>

#include <string.h>

static inline int next_char(unsigned char* buf, int size,
			    unsigned char*& cp, unsigned char*& end,
			    int fd) {
  if(cp == end) {
    int n = read(fd, buf, size);
    if(n <=0) {
      cp =buf;
      end = buf;
      return -1;
    } else {
      cp = buf;
      end = cp+n;
    }
  }
  return *cp++;
}

static inline extra_headers* cons_extra(char* field, char* value,
					extra_headers* rest) {
  extra_headers* tmp = new extra_headers;
  tmp->field = field;
  tmp->value = value;
  tmp->next = rest;
  return tmp;
}

static inline accept_val* cons_accept(char* value, accept_val* rest) {
  accept_val* tmp = new accept_val;
  tmp->value = value;
  tmp->next = rest;
  return tmp;
}

header* parse_header(unsigned char* buf, int size, int* offset,
		     int* valid, int fd)
// buf    ==> ptr to buffer
// size   ==> size of buffer
// offset <== returned value: offset of next byte in buf.
// valid  <=> returned value: number of valid bytes in buf. 
// fdet ==> fdet to read from if neccessary
{
  header* result = new header;
  memset(result,0,sizeof(header));
  
  unsigned char* cp = buf;
  unsigned char* end = buf+*valid;
  
  unsigned char* value=0;
  
  while(1) {
    int c = next_char(buf,size,cp,end,fd);
    
    if(c <0 ) { // error on fdet - exit
      goto error;
    }
    
    if(c == '\r') { 
      c = next_char(buf,size,cp,end,fd); // get rid of any <cr>s
      if(c <0) {
	goto done;
      }
    }
    if(c == '\n') { // we've hit an empty line - no more headers!
      goto done;
    }
    
    // we're at the start of a tag. Lets build up the tag
    char* tag =     new char[16];
    char* tag_ptr = tag;
    char* tag_end = tag+16;

    // add first char to tag buffer
    push_char(tag,tag_end,tag_ptr,c);

    // then loop until we hit a ':' or an error
    while(c = next_char(buf,size,cp,end,fd)) {
      if(c < 0 || c == '\r' || c == '\n') {
	delete tag;
	goto error;
      }
      if(c == ':') {
	break;
      } else {
	push_char(tag,tag_end,tag_ptr,c);
      }
    }
    // we've recieved a complete tag 
    push_char(tag,tag_end,tag_ptr,0);
    printf("tag is %s\n",tag);

    // get the space
    c = next_char(buf,size,cp,end,fd);
    if(c != ' ') {
      delete tag;
      goto error;
    }

    // get the value
    char* value =     new char[16];
    char* value_ptr = value;
    char* value_end = value+16;
    
    while(c = next_char(buf,size,cp,end,fd)) {
      if(c <0) {
	delete tag;
	delete value;
	goto error;
      }
      
      if(c == '\n') {
	break;
      }
      if(c != '\r') {
	push_char(value,value_end,value_ptr,c);
      }
    }
    // we're done with this value
    push_char(value,value_end,value_ptr,0);
    printf("value = %s\n",value);

    if(!strcmp(tag,"From")) {
      result->From = value;
      delete tag;
    } else if (!strcmp(tag,"Accept")) {
      result->Accept = cons_accept(value,result->Accept);
      delete tag;
    } else {
      // default
      result->extras = cons_extra(tag,value,result->extras);
    }
    
  }

  
 done: // we succeed in parsing the buffer ok
  *offset = cp - buf;
  *valid = end - buf;
  return( result);
 error: // we screwed up somewhere.
  printf("an error occured\n");
  return (0);
}

#ifdef TEST_PARSEHEADER

void test_parse() {
  unsigned char buf[1024];

  strcpy((char*)buf,"From: ses@unc.edu\nAccept: */*\r\n\r\n");
  int valid = strlen((char*)buf);
  int offset=0;
  header* foo = parse_header(buf,1024,&offset,&valid,-1);
  if(foo->From) {
    printf("from %s\n",foo->From);
  }
}
#endif
