#include "libmt.h"
#include "mt_daemon.h"
#include <errno.h>

#define log_error printf
int inet_daemon::setup() {
  set_priority(50);
  if(listen_sock <0) {
    listen_sock = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
    if(listen_sock <0) {
      log_error("inet_daemon: can't create socket(%s)\n",sys_errlist[errno]);
      return errno;
    }
    sockaddr_in sin;
    memset(&sin,0,sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port = port;

    int option =1;
    if(setsockopt(listen_sock,SOL_SOCKET,SO_REUSEADDR,(char*)&option,sizeof(int)) <0) {
      perror("setsockopt:");
    }
    if(bind(listen_sock,(sockaddr*)&sin,sizeof(sin)) <0) {
      log_error("listener_proc: can't bind socket to port %d (%s)\n",
		port, sys_errlist[errno]);
      close(listen_sock);
      return errno;
    }

    if(listen(listen_sock,64) <0) {
      log_error("listener_proc: can't listen to port %d (%d)\n",
		port,sys_errlist[errno]);
      close(listen_sock);
      return errno;
    }
  }
  return 0;
}

void inet_daemon::teardown() {
  if(listen_sock >= 0) {
    close(listen_sock);
  }
}

void* inet_daemon::get_connection() {
  sockaddr_in* whom;
  whom = new sockaddr_in;
  int* whom_size;
  whom_size = new int;
  *whom_size = sizeof(sockaddr_in);
  int new_fd;
 loop:
  new_fd=accept(listen_sock,(sockaddr*)whom,whom_size);
  
  if(new_fd <0 ) {
    if(errno == EMFILE) {
      log_error("out of file handles... sleeping\n");
      sleep(1);
      goto loop;
    } else {
      log_error("listener_proc: accept error (%s)\n",
		sys_errlist[errno]);
      delete whom_size;
      delete whom;
      return 0;
    }
  } else {
    return new inet_connect_info(new_fd,whom,whom_size);
  }
}
	
