#include <sys/types.h>
#include <sys/time.h>
#include <thread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>
#include <sysent.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <netdb.h>
#include "getopt.h"

extern "C" {
 int gettimeofday(struct timeval *tp);
}

// CLASS DEFINITIONS

// class request manages a list of filenames with their respective frequencies
class request {
public:
  char* filename;
  long frequency;
  request* next;

  request(char*fi, long freq, request* n) {
    filename = fi; frequency = freq; next = n;
  }
  ~request() {
    delete filename;
    if(next) delete next;
  }
  int length() {
    if(next == 0) {
      return 1;
    } else {
      return 1+next->length();
    }
  }
};

class frequency_node {
public:
  char* filename;
  long cumulative;
  frequency_node(char* f, long c) {
    filename = f; cumulative = c;
  }
  frequency_node() {
    filename = 0 ; cumulative = 0;
  }
};


class selector {
public:
  frequency_node* nodes;
  int num_nodes;
  long total_count;
  selector(request* requests) {
    total_count =0;
    num_nodes = requests->length();
    nodes = new  frequency_node[num_nodes];
    int i;
    for(i=0;i<num_nodes;i++) {
      total_count += requests->frequency;
      nodes[i].filename = requests->filename;
      nodes[i].cumulative = total_count;
      requests = requests->next;
    }
  }
  char* pick_name() {
    int key = int(rand() % total_count);
    int i;
    for(i=0;i<num_nodes;i++) {
      if (key < nodes[i].cumulative ) {
	return nodes[i].filename;
      }
    }
    printf("this shouldn't happen\n");
    return nodes[num_nodes-1].filename;
  }
};


request* read_requests(char* filename) {
  request* tmp = 0;
  
  FILE* req_file = fopen(filename,"r");
  if(req_file == 0) {
    printf("couldn't open %s\n",filename);
    return 0;
  }
  char request_name[1024];
  int frequency;
  
  while(fscanf(req_file,"%s %d\n",request_name,&frequency) != EOF) {
    tmp = new request(strdup(request_name),frequency,tmp);
  }

  return tmp;
}


// GLOBAL VARIABLES

hostent* he;
mutex_t result_lock;
mutex_t threads_left_lock;
mutex_t active_lock;
int active=0;
int waiting_for_connect=0;
int threads_left=0;
int result_wait=0;
double total_time;
long   total_bytes;
long   total_requests;

selector* random_names;
int port=80;
int verbose=0;

void* do_requests(void* arg);
// THE MAIN PROGRAM


main(int argc, char** argv) {
  
  char c;
  char* hostname="president";
  char* config_file="requests";
  int num_reps = 10;
  int num_threads = 1;
  
  while((c = getopt(argc,argv,"vh:p:f:r:n:R"))!= -1) {
    switch(c) {
    case 'v':
      verbose = 1;
      break;
    case 'h':
      hostname = strdup(optarg);
      break;
    case 'p':
      port = atoi(optarg);
      break;
    case 'f':
      config_file = strdup(optarg);
      break;
    case 'r':
      num_reps = atoi(optarg);
      break;
    case 'n':
      num_threads = atoi(optarg);
      break;
    case 'R':
      srand(int(getpid()));
      break;
    }
  }

  // look up hostname
  he = gethostbyname(hostname);
  if(he == 0) {
    printf("can't resolve hostname %s\n",hostname);
    exit(1);
  }
  
  request* requests = read_requests(config_file);
  
  if(requests == 0) {
    printf("bad request list file  %s\n",config_file);
    exit(1);
  }
// set up random filename selector
  random_names = new selector(requests);

// initialise locks

  mutex_init(&result_lock,USYNC_THREAD,0);
  mutex_init(&threads_left_lock,USYNC_THREAD,0);
  mutex_init(&active_lock,USYNC_THREAD,0);

  int i;
  mutex_lock(&threads_left_lock);
  threads_left = num_threads;
  for(i=0;i<num_threads;i++) {
    thr_create(NULL,0,do_requests,(void*)num_reps,THR_BOUND,NULL);
  }
  mutex_unlock(&threads_left_lock);
  pause();
}

void* do_requests(void* arg) {
  int num_requests = (int) arg;
  int i;
  sockaddr_in sin;

  memset(&sin,0,sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = port;
  memcpy(&sin.sin_addr,he->h_addr_list[0],he->h_length);
  char buf[8192];
  
  for(i=0;i<num_requests;i++) {
    int sock = socket(AF_INET,SOCK_STREAM,0);
    if(sock <0) {
      perror("couldn't create socket:");
      exit(1);
    }
    timeval start;
    timeval end;

    gettimeofday(&start);
    mutex_lock(&active_lock);
    waiting_for_connect++;
    mutex_unlock(&active_lock);
    if(connect(sock,(sockaddr*)&sin, sizeof(sin)) <0) {
	printf("errno = %d\n", errno); 
      perror("couldn't connect:");
    //  exit(1);
    }
    mutex_lock(&active_lock);
    active++;
    waiting_for_connect--;
    if(verbose) {
      printf("%d active, %d  connect, %d result_wait, %d total\n",
	     active, waiting_for_connect,result_wait,threads_left);
    }
    mutex_unlock(&active_lock);
    
    if(verbose) {
      gettimeofday(&end);
      if(end.tv_usec < start.tv_usec) {
	end.tv_sec -=1;
	end.tv_usec += 1000000;
      }
      printf("connect took %02.2lf seconds\n",
	     (double(end.tv_sec - start.tv_sec) +
	      (double(end.tv_usec - start.tv_usec) / 1000000.0)));
    }
    char* fname = random_names->pick_name();
    sprintf(buf,"GET %s\r\n",fname);
    write(sock,buf,strlen(buf));
    int n;
    int bytes_read=0;
    while((n= (read(sock,buf,8192))) >0) {
      bytes_read += n;
    }
    gettimeofday(&end);
    if(end.tv_usec < start.tv_usec) {
      end.tv_sec -=1;
      end.tv_usec += 1000000;
    }
    close(sock);
    mutex_lock(&active_lock);
    active--;
    result_wait++;
    mutex_unlock(&active_lock);
      
    mutex_lock(&result_lock);
    total_requests++;
    total_bytes += bytes_read;
    total_time += (double(end.tv_sec - start.tv_sec) +
			 (double(end.tv_usec - start.tv_usec) / 1000000.0));
    mutex_unlock(&result_lock);
    mutex_lock(&active_lock);
    result_wait--;
    mutex_unlock(&active_lock);
    if(verbose) {
      printf("%s: %d bytes in %02.2lf seconds\n",
	     fname,
	     bytes_read,
	     double(double(end.tv_sec - start.tv_sec) +
		    (double(end.tv_usec - start.tv_usec) / 1000000.0)));
    }
  }
// we're done - are we the last ones to finish?
  mutex_lock(&threads_left_lock);
  threads_left--;
  if(threads_left == 0) {
    printf("%d requests, %d bytes, %02.2lf seconds\n",
	   total_requests,total_bytes,total_time);
    printf("average size, %3.2lf bytes, average time %02.2lf seconds\n",
	   double(total_bytes)/double(total_requests),
	   double(total_time)/double(total_requests));
    exit(0);
  } else {
    mutex_unlock(&threads_left_lock);
  }
  return 0;
}
